#!/usr/bin/env python

import argparse
import sys

from rna import RNA
from sub_command import SubCommand as cmd


class RNA2d(object):
    def __init__(self):
        self.__version__ = "1.0.0"
        self.formats = RNA.ss_formats
        self.helps = {'format-input': 'Format of ss input, can be ' + ','.join(self.formats),
                      'format-output': 'Format of ss output can be ' + ','.join(self.formats)}
        self.parser = argparse.ArgumentParser(prog='2dRNA',
                                              description="This program is used to expand the base pairs after removing.")
        self.parser.add_argument('--version', action='version', version=self.__version__)
        self.add_subparsers()

    def add_subparsers(self):
        self.subparsers = self.parser.add_subparsers(
            dest='subcommand', help='Available subcommands')
        self.parser_expand = self.subparsers.add_parser(
            "expand", help="Expand the DI predicted pairs.")
        self.parser_filter = self.subparsers.add_parser(
            "filter", help="Filter the DI results.")
        self.parser_bench = self.subparsers.add_parser(
            "bench", help="Bench mark the DI results.")
        self.parser_fold = self.subparsers.add_parser(
            "fold", help="Fold the given RNA sequence"
        )
        self.parser_format = self.subparsers.add_parser(
            "format", help="transform the RNA secondary structure to other formats"
        )

    def add_arguments(self):
        self.add_fold_arguments()
        self.add_filter_arguments()
        self.add_expand_arguments()
        self.add_bench_arguments()
        self.add_format_arguments()

    def add_format_arguments(self):
        self.parser_format.add_argument("-i", "--ss", type=str, required=True,
                                        metavar="[RNA SS]", dest='ss_file',
                                        help='Source RNA SS file')
        self.parser_format.add_argument("-o", "--output", type=str, required=False,
                                        metavar="[OUTPUT]", dest="output",
                                        help="Output file name")
        self.parser_format.add_argument("-of", "--out-fmt", type=str, required=False,
                                        choices=self.formats,
                                        default="dbn", metavar="[FORMAT]",
                                        dest="out_fmt", help="Format of target")
        self.parser_format.add_argument("-if", "--ss-fmt", type=str, required=False,
                                        choices=self.formats, metavar="[FORMAT]",
                                        dest="ss_fmt", help="Format of target")
        self.parser_format.add_argument("-s", "--seq", type=str, metavar="[RNA SEQUENCE]",
                                        dest='seq_file', help='RNA sequence file')
        self.parser_format.add_argument("--no-pk", action="store_true",
                                        default=False, dest="no_pk",
                                        help="Remove pseudo-knots from input")
        self.parser_format.set_defaults(func=cmd.format)

    def add_fold_arguments(self):
        self.parser_fold.add_argument("-s", "--seq", type=str, required=False,
                                      metavar="[SEQUENCE FILE]", dest='seq_file', help="sequence file in fasta format")
        self.parser_fold.add_argument("-i", "--ss", type=str, required=True,
                                      metavar="[SS FILE]", dest='ss_file', help="ss file name")
        self.parser_fold.add_argument("-if", "--ss-fmt", type=str, required=False, choices=self.formats,
                                      metavar="[SS FORMAT]", dest='ss_fmt',
                                      help=self.helps['format-input'])
        self.parser_fold.add_argument("-m", "--method", type=str, required=True,
                                      choices=["mfold", "RNAfold", "RNAstructure"], metavar="[METHOD]",
                                      dest="method", help="The folding method: mfold, RNAfold or RNAstructure")
        self.parser_fold.add_argument("-e", "--exe", type=str, required=False,
                                      metavar="[EXE PATH]", dest='exe',
                                      help="The path of fold executable")
        self.parser_fold.add_argument("-o", "--output", type=argparse.FileType('w'), default=sys.stdout,
                                      metavar="[OUTPUT]", dest='output',
                                      help="the final output")
        self.parser_fold.add_argument("-of", "--out-fmt", type=str, required=False, default='dbn',
                                      choices=self.formats,
                                      metavar="[SS FORMAT]", dest='out_fmt',
                                      help=self.helps['format-output'])
        self.parser_fold.set_defaults(func=cmd.fold)

    def add_filter_arguments(self):
        self.parser_filter.add_argument("--di", type=argparse.FileType('r'), required=True,
                                        metavar="[DI FILE]", dest='di_file', help="DI file")
        self.parser_filter.add_argument("--seq", type=argparse.FileType('r'), required=True,
                                        metavar="[SEQUENCE FILE]", dest='seq_file', help="Sequence file")
        self.parser_filter.add_argument("--top-n", type=str, default="L", metavar="[TOP N]",
                                        dest="topn", help="Top N of sequence, only for DCA input")
        self.parser_filter.add_argument("--pre-clean", action="store_true",
                                        dest="clean", help="Clean DCA result before filter.")
        self.parser_filter.add_argument("-o", "--output", type=argparse.FileType('w'), default=sys.stdout,
                                        metavar="[OUTPUT FILE]", dest='output', help='Output file')
        self.parser_filter.add_argument("-of", "--output-fmt", type=str, default='di',
                                        choices=self.formats,
                                        metavar="[FORMAT STRING]",
                                        dest='out_fmt', help=self.helps['format-output'])
        self.parser_filter.add_argument("--no-pk", action="store_true",
                                        default=False, dest="no_pk",
                                        help="Remove pseudo-knots from input")
        self.parser_filter.add_argument("--ml", action="store_true", dest="is_ml",
                                        help="The input is from machine-learning based method?")
        self.parser_filter.add_argument("--cutoff", type=float, default=0.5, metavar="[CUTOFF]",
                                        dest="cutoff", help="Probility cutoff for machine-learning based method.")
        self.parser_filter.set_defaults(func=cmd.filter)

    def add_expand_arguments(self):
        self.parser_expand.add_argument("--seq", type=argparse.FileType('r'), required=False, default='',
                                        metavar="[SEQUENCE FILE]", dest='seq_file', help="sequence file name")
        self.parser_expand.add_argument("--ss", type=argparse.FileType('r'), required=True,
                                        metavar="[SS FILE]", dest='ss_file', help='SS file name')
        self.parser_expand.add_argument("-o", "--output", type=argparse.FileType('w'), default=sys.stdout,
                                        metavar="[OUTPUT]", dest='output', help='Output file')
        self.parser_expand.add_argument("-if", "--ss-fmt", type=str, metavar="[FORMAT STRING]",
                                        choices=self.formats, dest='ss_fmt',
                                        help=self.helps['format-input'])
        self.parser_expand.add_argument("-of", "--output-fmt", type=str, default='dbn',
                                        metavar="[FORMAT STRING]", choices=self.formats,
                                        dest='out_fmt', help=self.helps['format-output'])
        self.parser_expand.add_argument("--no-pk", action="store_true",
                                        default=False, dest="no_pk",
                                        help="Remove pseudo-knots from input")
        self.parser_expand.set_defaults(func=cmd.expand)

    def add_bench_arguments(self):
        self.parser_bench.add_argument("-i", "--predict-file", type=argparse.FileType('r'), required=True,
                                       metavar="[PREDICTION]", dest='predict_file', help='DI file or prediction')
        self.parser_bench.add_argument("-n", "--native-file", type=argparse.FileType('r'), required=True,
                                       metavar="[NATIVE PAIR]", dest='native_file', help='Native base pairs')
        self.parser_bench.add_argument("-if", "--predict-fmt", type=str, required=False,
                                       choices=self.formats,
                                       metavar="[PREDICT FILE FORMAT]", dest='predict_fmt',
                                       help='Format of prediction input file')
        self.parser_bench.add_argument("-nf", "--native-fmt", type=str, required=False,
                                       choices=self.formats,
                                       metavar="[NATIVE FILE FORMAT]", dest='native_fmt',
                                       help='Format of native input file')
        self.parser_bench.add_argument("-s", "--seq-file", type=argparse.FileType('r'), required=False,
                                       metavar="[SEQUENCE FILE]", dest='seq_file', help="Sequence file")
        self.parser_bench.set_defaults(func=cmd.bench)

    def __call__(self):
        self.add_arguments()
        args = self.parser.parse_args()
        args.func(args)


if __name__ == "__main__":
    rna2d = RNA2d()
    rna2d()
