import string

from error import RNA2dError
from ss_parser import SSParser


class DBNParser(SSParser):
    def __init__(self):
        super(DBNParser, self).__init__()
        self.seq_symbols = ['A', 'U', 'G', 'C', 'T']
        self.ss_symbol_map = {
            ')': '(',
            ']': '[',
            '}': '{',
        }
        for c in string.ascii_lowercase:
            self.ss_symbol_map[c] = c.upper()
        self.ss_left_symbols = list(self.ss_symbol_map.values())
        self.ss_right_symbols = list(self.ss_symbol_map.keys())
        self.all_symbols = ['.'] + self.ss_left_symbols + self.ss_right_symbols

    def _parse_dbn(self, line):
        line = line.strip().split()[0]
        pairs = []
        stacks = dict()
        for c in self.ss_left_symbols:
            stacks[c] = []
        for idx, c in enumerate(line, 1):
            if c not in self.all_symbols:
                raise RNA2dError('Invalid character in secondary structure')
            if c in self.ss_left_symbols:
                stacks[c].append(idx)
            elif c in self.ss_right_symbols:
                left_symbol = self.ss_symbol_map[c]
                if (len(stacks[left_symbol]) == 0):
                    raise RNA2dError('Invalid DBN string, unpaired bracket!')
                resi = stacks[left_symbol].pop()
                pairs.append((resi, idx))
        for value in stacks.values():
            if len(value) != 0:
                raise RNA2dError('Invalid DBN string, unpaired bracket!')
        return pairs

    def _parse_seq_ss(self, input_file, strip=0):
        with open(input_file, 'r') as f:
            for i in range(strip):
                _ = f.readline()
            lines = f.readlines()
        return self._parse_seq_ss_str("".join(lines), strip)

    def _parse_seq_ss_str(self, input_str, strip=0):
        lines = input_str.strip().split('\n')[strip:]
        seq, ss_line = lines
        pairs = self._parse_dbn(ss_line)
        return seq, pairs

    def parse_str(self, input_str):
        if input_str.startswith('>'):
            seq, pairs = self._parse_seq_ss_str(input_str, strip=1)
        elif input_str[0].upper() in self.seq_symbols:
            seq, pairs = self._parse_seq_ss_str(input_str, strip=0)
        else:
            seq = None
            pairs = self._parse_dbn(input_str)
        return seq, pairs

    def parse(self, input_file):
        with open(input_file, 'r') as f:
            lines = f.readlines()
        return self.parse_str("".join(lines))


if __name__ == "__main__":
    import sys

    file = sys.argv[1]
    parser = DBNParser()
    seq, pairs = parser.parse(file)
    print(seq)
    print(pairs)
    print(len(pairs))
