from di_parser import DIParser
from error import RNA2dError
from factory import Factory
from rna import RNA
from seq_parser import SeqParser
from ss_transformer import SSTransformer
import math


class DIFilter(object):
    def __init__(self, args):
        self.args = args
        self.output = args.output
        self.out_fmt = args.out_fmt
        self.no_pk = args.no_pk
        self.di_file = args.di_file.name
        self.seq_file = args.seq_file.name
        self.conv_pairs = RNA.conv_pairs
        self.sequence = self._read_seq_file()
        self.size = len(self.sequence)
        self.filtered_di_values = self._read_di_file()
        self.is_ml = args.is_ml
        if args.topn:
            self.top_n = self.get_top_n(args.topn)
        if args.cutoff:
            self.cutoff = args.cutoff
        self.clean = args.clean
        self.conv_di_values = []

    def get_top_n(self, top_n_str):
        if top_n_str == "L":
            top_n = self.size
        elif top_n_str.startswith("L"):
            top_n = self.size / int(top_n_str[1:])
        elif top_n_str.endswith("L"):
            top_n = int(float(top_n_str[:-1]) * self.size)
        else:
            try:
                top_n = int(top_n_str)
            except Exception:
                raise RNA2dError("Invalid top_n value, should be xL or Lx.")
        if top_n < 1:
            top_n = 1
        return top_n

    def _read_di_file(self):
        parser = DIParser()
        _, di_values = parser.parse(self.di_file)
        return di_values

    def _read_seq_file(self):
        parser = SeqParser()
        self.sequence = parser.parse(self.seq_file)
        return self.sequence

    def _remove_non_conv(self):
        result = []
        for resi, resj, di_value in self.filtered_di_values:
            resname_i = self.sequence[resi - 1]
            resname_j = self.sequence[resj - 1]
            if (resname_i, resname_j) in self.conv_pairs:
                result.append((resi, resj, di_value))
        self.filtered_di_values = result
        self.conv_di_values = result

    def _extract_top_n(self):
        if self.top_n <= len(self.filtered_di_values):
            self.filtered_di_values = self.filtered_di_values[:self.top_n]
        else:
            self.filtered_di_values = self.filtered_di_values

    def _extract_by_cutoff(self):
        result = []
        for i,j,v in self.filtered_di_values:
            if v >= self.cutoff:
                result.append((i, j, v))
        self.filtered_di_values = result

    def _remove_conflict2(self):
        result = []
        residue_count = dict()
        for resi, resj, di_value in self.filtered_di_values:
            if resi in residue_count:
                residue_count[resi] += 1
            else:
                residue_count[resi] = 1
            if resj in residue_count:
                residue_count[resj] += 1
            else:
                residue_count[resj] = 1
        for resi, resj, di_value in self.filtered_di_values:
            if residue_count[resi] == 1 and residue_count[resj] == 1:
                result.append((resi, resj, di_value))
        self.filtered_di_values = result

    def _resid_pairs_map(self):
        result = dict()
        for idx,record in enumerate(self.filtered_di_values):
            resi, resj, di_value = record
            if resi in result:
                result[resi].append(idx)
            else:
                result[resi] = [idx]
            if resj in result:
                result[resj].append(idx)
            else:
                result[resj] = [idx]
        return result

    def _get_up_pairs(self, pair):
        resi, resj, _ = pair
        result = 0
        while resi <= self.size and resj > 0:
            if self._is_possible_paired(resi, resj):
                result += 1
            else:
                break
            resi += 1
            resj -= 1
        return result

    def _get_down_pairs(self, pair):
        resi, resj, _ = pair
        result = 0
        while resi > 0 and resj <= self.size:
            if self._is_possible_paired(resi, resj):
                result += 1
            else:
                break
            resi -= 1
            resj += 1
        return result

    def _get_possible_pairs(self, pair):
        return self._get_up_pairs(pair) + self._get_down_pairs(pair) - 1

    def _is_most_likely_paired(self, pair, resid_pairs):
        resi, resj, _ = pair
        n_of_possible_pairs = self._get_possible_pairs(pair)
        conflicted_pair_idx = set(resid_pairs[resi] + resid_pairs[resj])
        for idx in conflicted_pair_idx:
            t_possible_pairs = self._get_possible_pairs(self.filtered_di_values[idx])
            if t_possible_pairs > n_of_possible_pairs:
                return False
        return True

    def _is_conflict_with_previous(self, pair, all_pairs):
        all_idx = set([p[0] for p in all_pairs] + [p[1] for p in all_pairs])
        resi, resj, _ = pair
        if resi in all_idx or resj in all_idx:
            return True
        else:
            return False

    def _remove_conflict(self):
        result = []
        resid_pairs = self._resid_pairs_map()
        for idx,pair in enumerate(self.filtered_di_values):
            resi, resj, di_value = pair
            if len(resid_pairs[resi]) == 1 and len(resid_pairs[resj]) == 1:
                result.append((resi, resj, di_value))
            elif self._is_most_likely_paired(pair, resid_pairs):
                if not self._is_conflict_with_previous(pair, result):
                    result.append((resi, resj, di_value))
        self.filtered_di_values = result

    def _get_neighbors(self, resi, resj, gap):
        neighbors = dict()
        for i in range(1, gap + 1):
            resi_prev = resi - i
            resi_next = resi + i
            resj_prev = resj + i
            resj_next = resj - i
            if resi_prev > 0 and resj_prev <= self.size:
                neighbors[-i] = (resi_prev, resj_prev)
            if resi_next <= self.size and resj_next > 0:
                neighbors[i] = (resi_next, resj_next)
        return neighbors

    def _is_possible_paired(self, resi, resj):
        resn_i = self.sequence[resi - 1]
        resn_j = self.sequence[resj - 1]
        if resj - resi > 3 and (resn_i, resn_j) in self.conv_pairs:
            return True
        else:
            return False

    def _count_false(self, values):
        count = 0
        for v in values:
            if not v:
                count += 1
        return count

    def _is_paired(self, pair):
        conv_pairs = [(i, j) for i,j,_ in self.filtered_di_values]
        resi, resj = pair
        if (resi, resj) in conv_pairs or (resj, resi) in conv_pairs:
            return True
        else:
            return False

    def _at_least_three_continuous_pairs(self, is_not_paired):
        # -1 0 1
        if not is_not_paired[-1] and not is_not_paired[1]:
            return True
        # -2 -1 0
        elif not is_not_paired[-2] and not is_not_paired[-1]:
            return True
        # 0 1 2
        elif not is_not_paired[1] and not is_not_paired[2]:
            return True
        else:
            return False

    def _at_least_two_continuous_pairs(self, is_not_paired):
        # -1 0
        if not is_not_paired[-1]:
            return True
        # 0  1
        elif not is_not_paired[1]:
            return True
        else:
            return False

    def _is_not_alone(self, is_not_paired, pair, gap=2):
        for i in range(1, gap + 1):
            if i not in is_not_paired:
                is_not_paired[i] = True
            if -i not in is_not_paired:
                is_not_paired[-i] = True
        resi, resj = pair
        next_resi, next_resj = resi + 1, resj - 1
        prev_resi, prev_resj = resi - 1, resj + 1
        if self._is_paired((next_resi, next_resj)) \
            or self._is_paired((prev_resi, prev_resj)):
            return True
        else:
            if self.is_ml:
                return self._at_least_two_continuous_pairs(is_not_paired)
            else:
                return self._at_least_three_continuous_pairs(is_not_paired)

    def _conflict_with_bps(self, resi, resj):
        pair_map = {resi: resj for resi, resj, _ in self.filtered_di_values}
        if resi > resj:
            resi, resj = resj, resi
        if ((resi in pair_map) and pair_map[resi] != resj) or \
                ((resj in pair_map) and pair_map[resj] != resi):
            return True
        else:
            return False

    def _remove_single_points(self, gap=2, second=False):
        result = []
        for resi, resj, di_value in self.filtered_di_values:
            is_not_paired = dict()
            neighbors = self._get_neighbors(resi, resj, gap)
            for key, value in neighbors.items():
                t_resi, t_resj = value
                if not self._is_possible_paired(t_resi, t_resj):
                    is_not_paired[key] = True
                else:
                    if second and self._conflict_with_bps(t_resi, t_resj):
                        is_not_paired[key] = True
                    else:
                        is_not_paired[key] = False
            if self._is_not_alone(is_not_paired, (resi, resj), gap):
                result.append((resi, resj, di_value))
        self.filtered_di_values = result

    def write(self):
        self.filtered_di_values.sort(key=lambda x: x[0])
        pairs = [(pair[0], pair[1]) for pair in self.filtered_di_values]
        di_values = [p[2] for p in self.filtered_di_values]
        data = {'seq': self.sequence,
                'pairs': pairs,
                'di': di_values}
        writer = Factory.make_ss_writer(self.out_fmt)
        writer.write(data, self.output)

    def _prepare(self):
        if self.clean:
            self._remove_non_conv()
            if self.is_ml:
                self._extract_by_cutoff()
            else:
                self._extract_top_n()
        else:
            if self.is_ml:
                self._extract_by_cutoff()
            else:
                self._extract_top_n()
            self._remove_non_conv()

    def _remove_pk(self):
        self.filtered_di_values, _ = SSTransformer.remove_pk(self.filtered_di_values)

    def run(self):
        self._prepare()
        self._remove_single_points()
        self._remove_conflict()
        self._remove_single_points(second=True)
        if self.no_pk:
            self._remove_pk()
