import os
import shutil
import tempfile
from abc import abstractmethod

from error import RNA2dError
from helper import get_file_ext
from seq_parser import SeqParser
from seq_writer import SeqWriter


class Folder(object):
    def __init__(self, args):
        self.args = args
        self.ss_file = self.args.ss_file
        self.ss_fmt = self.get_ss_fmt()
        self.out_fmt = args.out_fmt
        self.seq, self.pairs = self.get_seq_pairs()
        self.output = self.args.output
        self.temp_seq = tempfile.mktemp()
        self.temp_cstr = tempfile.mktemp()
        self.command = ""
        self.path = ""

    def get_ss_fmt(self):
        if self.args.ss_fmt:
            return self.args.ss_fmt
        else:
            return get_file_ext(self.args.ss_file)

    def get_seq_pairs(self):
        from factory import Factory
        parser = Factory.make_ss_parser(self.ss_fmt)
        seq, pairs = parser.parse(self.ss_file)
        if not seq:
            if not self.args.seq_file:
                raise RNA2dError('The sequence file is needed!')
            else:
                parser = SeqParser()
                seq = parser.parse(self.args.seq_file)
        return seq, pairs

    @staticmethod
    def add_pk_to_pairs(pairs, pk_pairs):
        paired_bases = set([bp[0] for bp in pairs] + [bp[1] for bp in pairs])
        pk_bps = []
        for pk in pk_pairs:
            i, j, n = pk
            for k in range(n):
                pk_bps.append((i + k, j - k))
        for res_i, res_j in pk_bps:
            if res_i not in paired_bases and res_j not in paired_bases:
                pairs.append((res_i, res_j))
        pairs.sort(key=lambda x: x[0])
        return pairs

    def __del__(self):
        if os.path.exists(self.temp_seq):
            os.remove(self.temp_seq)
        if os.path.exists(self.temp_cstr):
            os.remove(self.temp_cstr)

    def write_seq(self):
        writer = SeqWriter()
        writer.write(self.seq, self.temp_seq)

    def write_out(self, pairs):
        from factory import Factory
        data = {'seq': self.seq,
                'pairs': pairs}
        writer = Factory.make_ss_writer(self.out_fmt)
        writer.write(data, self.output)

    def set_command(self, cmd_name):
        self.command = cmd_name
        self.path = shutil.which(self.command)
        if not self.path:
            if not self.args.exe:
                raise RNA2dError('Cannot find the executable: ' + cmd_name)
            else:
                path = os.path.realpath(self.args.exe)
                if not os.path.exists(path):
                    raise RNA2dError('Cannot find the executable: ' + cmd_name)
                self.command = path
        else:
            if not self.args.exe:
                self.command = self.path
            else:
                self.command = os.path.realpath(self.args.exe)

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def write_cstr(self, data):
        pass
