#!/usr/bin/env python

import sys

from error import RNA2dError
from factory import Factory
from helper import get_file_ext
from rna import RNA


class Perf(object):
    def __init__(self, args):
        self.args = args
        self.formats = RNA.ss_formats
        self._set_params()
        self._read_files()
        self._init()

    def _set_predict_fmt(self):
        if self.args.predict_fmt:
            self.predict_fmt = self.args.predict_fmt
        else:
            self.predict_fmt = get_file_ext(self.predict_file)

    def _set_native_fmt(self):
        if self.args.native_fmt:
            self.native_fmt = self.args.native_fmt
        else:
            self.native_fmt = get_file_ext(self.native_file)

    def _set_seq(self):
        if self.predict_fmt == 'di' and self.native_fmt == 'di':
            if not self.args.seq_file:
                raise RNA2dError("The sequence file is needed when the predict "
                                 "file and native file is both in 'di' format")
            else:
                self.seq_file = self.args.seq_file.name

    def _set_params(self):
        self.predict_file = self.args.predict_file.name
        self.native_file = self.args.native_file.name
        self._set_predict_fmt()
        self._set_native_fmt()
        self._set_seq()

    def _read_files(self):
        self._read_seq_file()
        self._read_predict_file()
        self._read_native_file()

    @staticmethod
    def strip_di(pairs):
        return [(p[0], p[1]) for p in pairs]

    def _read_seq_file(self):
        from seq_parser import SeqParser
        if not self.args.seq_file:
            return
        parser = SeqParser()
        self.seq = parser.parse(self.args.seq_file.name)
        self.seq_len = len(self.seq)

    def _read_native_file(self):
        parser = Factory.make_ss_parser(self.native_fmt)
        seq, self.native_pairs = parser.parse(self.native_file)
        self.native_pairs = Perf.strip_di(self.native_pairs)
        if hasattr(self, 'seq') and seq and self.seq != seq:
            raise RNA2dError("The native sequence and predic sequence is not paired!")
        self.seq = seq

    def _read_predict_file(self):
        parser = Factory.make_ss_parser(self.predict_fmt)
        seq, self.predict_pairs = parser.parse(self.predict_file)
        self.predict_pairs = Perf.strip_di(self.predict_pairs)
        if hasattr(self, 'seq') and seq and self.seq != seq:
            raise RNA2dError("The native sequence and predic sequence is not paired!")
        self.seq = seq

    def get_tp(self):
        count = 0
        for pair in self.predict_pairs:
            if pair in self.native_pairs:
                count += 1
        return count

    def get_fp(self):
        count = 0
        for pair in self.predict_pairs:
            if pair not in self.native_pairs:
                count += 1
        return count

    def get_tn(self):
        result = (self.seq_len * (self.seq_len - 1)) / 2
        result = result - (self.TP + self.FP + self.FN)
        return result

    def get_fn(self):
        return len(self.native_pairs) - self.TP

    def _init(self):
        self.TP = self.get_tp()
        self.FP = self.get_fp()
        self.FN = self.get_fn()
        self.TN = self.get_tn()

    def calc_sty(self):
        return self.TP * 1.0 / len(self.native_pairs)

    def calc_ppv(self):
        return self.TP * 1.0 / len(self.predict_pairs)

    def calc_mcc(self):
        mcc = (self.TP * self.TN - self.FP * self.FN) * 1.0 / \
              ((self.TP + self.FP) * (self.FP + self.TN) * ( \
                          self.TN + self.FN) * (self.TP + self.FN)) ** 0.5
        return mcc

    def bench(self):
        fmt = "{:s} {:8.4f} {:8.4f} {:8.4f}\n"
        sys.stdout.write(fmt.format(self.predict_file,
                                    self.calc_ppv(),
                                    self.calc_sty(),
                                    self.calc_mcc()))
