#!/usr/bin/env python

from error import RNA2dError
from factory import Factory
from rna import RNA
from seq_parser import SeqParser


class SSExpander(object):
    def __init__(self, args):
        self.args = args
        self.no_pk = args.no_pk
        self.seq_file = args.seq_file.name
        self.ss_file = args.ss_file.name
        self.min_loop = 4
        self.ss_fmt = args.ss_fmt
        self.output = args.output
        self.out_fmt = args.out_fmt
        self.conv_pairs = RNA.conv_pairs
        self.seq, self.pairs = self._read_file()
        self.resid_in_input = [p[0] for p in self.pairs] + [p[1] for p in self.pairs]
        self.size = len(self.seq)
        self.expanded_pairs = [p for p in self.pairs]

    def _read_file(self):
        parser = Factory.make_ss_parser(self.ss_fmt)
        seq, pairs = parser.parse(self.ss_file)
        if self.ss_fmt == 'di':
            pairs = [(p[0], p[1]) for p in pairs]
        if not seq:
            if self.seq_file == '':
                raise RNA2dError('The sequence is needed, please provide a sequence file')
            else:
                parser = SeqParser()
                seq = parser.parse(self.seq_file)
        return seq, pairs

    def _is_overflow(self, resi, resj):
        return resi > self.size or resj > self.size or resi < 1 or resj < 1

    def _gap_is_too_small(self, resi, resj):
        return abs(resi - resj) <= self.min_loop

    def _is_conflict_with_input(self, resi, resj):
        if resi in self.resid_in_input or resj in self.resid_in_input:
            return True
        else:
            return False

    def _expand_pair(self, pair, strip=1):
        resi, resj = pair
        while True:
            resi += strip
            resj -= strip
            if self._is_overflow(resi, resj) \
                    or self._gap_is_too_small(resi, resj)\
                    or self._is_conflict_with_input(resi, resj):
                break
            resname_i = self.seq[resi - 1]
            resname_j = self.seq[resj - 1]
            if (resname_i, resname_j) in self.conv_pairs:
                self.expanded_pairs.append((resi, resj))
            else:
                break

    def _expand(self):
        for pair in self.expanded_pairs:
            self._expand_pair(pair, 1)
        for pair in self.expanded_pairs:
            self._expand_pair(pair, -1)
        tmp = set(self.expanded_pairs)
        self.expanded_pairs = sorted(list(tmp), key=lambda x: x[0])

    def _score_left(self, pair):
        score = 0
        resi, resj = pair
        while True:
            prev_pair = (resi - 1, resj + 1)
            if prev_pair in self.expanded_pairs:
                score += 1
                resi, resj = prev_pair
            else:
                break
        return score

    def _score_right(self, pair):
        score = 0
        resi, resj = pair
        while True:
            next_pair = (resi + 1, resj - 1)
            if next_pair in self.expanded_pairs:
                score += 1
                resi, resj = next_pair
            else:
                break
        return score

    def _score(self, pair):
        if pair[0] > pair[1]:
            pair = (pair[1], pair[0])
        score = self._score_left(pair) + self._score_right(pair)
        return score

    def _is_pairs_conflict(self, pair1, pair2):
        for i in pair1:
            if i in pair2:
                return True
        return False

    def _remove_conflict(self):
        result = []
        n_pairs = len(self.expanded_pairs)
        idx_2b_removed = []
        for i in range(n_pairs-1):
            pair_i = self.expanded_pairs[i]
            for j in range(i+1, n_pairs):
                pair_j = self.expanded_pairs[j]
                if self._is_pairs_conflict(pair_i, pair_j):
                    if pair_i in self.pairs and pair_j not in self.pairs:
                        idx_2b_removed.append(j)
                    elif pair_i not in self.pairs and pair_j in self.pairs:
                        idx_2b_removed.append(i)
                    elif pair_i not in self.pairs and pair_j not in self.pairs:
                        idx_2b_removed.append(i)
                        idx_2b_removed.append(j)
        for i in range(n_pairs):
            if i not in idx_2b_removed:
                result.append(self.expanded_pairs[i])
        self.expanded_pairs = result

    def _is_possibly_paired(self, pair):
        i, j = pair
        if i < 1 or j > self.size:
            return False
        if pair in self.expanded_pairs:
            return True
        res_i, res_j = self.seq[i-1], self.seq[j-1]
        if (res_i, res_j) in self.conv_pairs:
            return True
        else:
            return False

    def _remove_single(self):
        result = []
        for resi, resj in self.expanded_pairs:
            pair_prev = (resi - 1, resj + 1)
            pair_next = (resi + 1, resj - 1)
            if not self._is_possibly_paired(pair_prev) and \
                not self._is_possibly_paired(pair_next):
                continue
            else:
                result.append((resi, resj))
        self.expanded_pairs = result

    def run(self):
        from ss_transformer import SSTransformer
        self._expand()
        self._remove_conflict()
        self._remove_single()
        if self.no_pk:
            self.expanded_pairs, _ = SSTransformer.remove_pk(self.expanded_pairs)

    def write(self):
        data = {'seq': self.seq,
                'pairs': self.expanded_pairs}
        writer = Factory.make_ss_writer(self.out_fmt)
        writer.write(data, self.output)
